/*
 codec.C
 ------------------------------------------------------------------------
 A command-line compiler for CODE (strictly speaking, for a textual
 version of the CODE language).
 ------------------------------------------------------------------------
 @(#) $Id: codec.C,v 1.1 1999/11/09 20:18:32 emery Exp $
 ------------------------------------------------------------------------
 AUTHOR/CONTACT:
 
 Emery Berger                    | <http://www.cs.utexas.edu/users/emery>
 Parallel Programming Group      |  <http://www.cs.utexas.edu/users/code>
 Department of Computer Sciences |             <http://www.cs.utexas.edu>
 University of Texas at Austin   |                <http://www.utexas.edu>
 ========================================================================
*/


#include <config.h>

#include <assert.h>
#include <ctype.h>
#include <stream.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include <misc/mstring.h>
#include <misc/fileutil.h>

#include <misc/general.h>
#include <exmodel/abstree.h>
#include <exmodel/exmodel.h>
#include <cplparser/parser.h>

static char const rcsid[] = "$Id: codec.C,v 1.1 1999/11/09 20:18:32 emery Exp $";

int C2_ErrorUID;		// Stores the UID of an erroneous object.
void Die(char *Message);	// Print Message, clean up, and quit.

#include <misc/sfile.h>
#include <exmodel/object.h>
#include <symtab/stype.h>

#include <lib/Syntax.h>
#include <lib/UidTable.h>
#include <lib/MsgUI.h>
#include <translate/translate.h>

extern unsigned nextUID (void);


// Global variables used by the main program

cProgram *ProgramRoot;
cGraph *ThisGraph;
cUC * ThisUC;
MString FilesToLink;
TranslateType TranslatorT;
TraceLevel_t TraceLevel;
int OptimizeProg = 1;
int ProduceSummaryStats = 1;
int AddDbgrCode = 0;

cBlock *cCursor::Root = 0;
cSType *cSType::Head = 0;


// Target types and associated strings.

enum { NUM_TARGETS = 5 };
const char * targets[] = { "mpi", "pvm", "smp", "serial", "list" };
TranslateType targetTypes[] = { MPI, PVM, UNIFIED, SERIAL, LIST };


// Print out a usage message.

static void printUsage (void)
{
  cout << "Usage: codec [options] FILE" << endl
       << "Options:" << endl
       << "  -h               Show this summary of options." << endl
       << "  -t TARGET        Select a compilation target (MPI, PVM, Serial, SMP)." << endl
       << "  -v               Show version information." << endl
       << endl;
}


#include <cplparser/scan.h>
extern int yyparse (void);

int main(int argc,
	 char *argv[])
{
  TestDirectory = new char[255];
  sprintf (TestDirectory, ".");

  char filename[255];

  // Parse the command line.
  extern char *optarg;
  int c;
  int i;
  char target[255];
  char lctarget[255];

  target[0] = '\0';
  filename[0] = '\0';

  while ( (c = getopt (argc, argv, "c:ht:v-")) != EOF ) {
    switch (c) {
    case 'h':
      // Help.
      printUsage ();
      exit (0);
      break;
    case 't':
      sprintf (target, "%s", optarg);
      // Select a particular target.
      // First, convert to lower-case.
      for (i = 0; i < strlen(target)+1; i++) {
	lctarget[i] = tolower (target[i]);
      }
      // Now look for a match.
      for (i = 0; i < NUM_TARGETS; i++) {
	if (strcmp (lctarget, targets[i]) == 0) {
	  TranslatorT = targetTypes[i];
	  break;
	}
      }
      // Alert the user if we didn't find a match.
      if (i == NUM_TARGETS) {
	cerr << "Error: " << target << " is not an available target." << endl;
	exit (1);
      }
      break;
    case 'c':
      // Compile a given file to the default target and exit.
      sprintf (filename, "%s", optarg);
      break;
    case 'v':
      // Version information.
      cout << "CODE 2.3" << endl
	   << "The CODE Visual Parallel Programming System" << endl
	   << "by Emery Berger, Peter Newton, Rajeev Vokkarne and Brian West." << endl
	   << "Copyright (C) 1999 The University of Texas at Austin." << endl
	   << "See http://www.cs.utexas.edu/users/code for more information." << endl
	   << endl;
      exit (0);
      break;
    case '-':
      goto doneparsing;
    }
  }
  
  // Complain and quit if we haven't
  // got both a target and a filename.
  if (!strcmp (target, "") || !strcmp (filename, "")) {
    cerr << "Error: You need to specify a target and a filename." << endl;
    exit (1);
  }

doneparsing:
   
  //
  // Read the whole input file into a string.
  //

  char buf[65536]; // Arbitrary 64K limit! FIX ME!!
  FILE * f = fopen (filename, "r");
  char * bufptr = &buf[0];

  if (!f) {
    cerr << "Error: Could not open file \"" << filename << "\"." << endl;
    exit (1);
  }

  size_t sz;

  while ((sz = fread ((void *) bufptr, 1, 65536, f))) {
    bufptr += sz;
  }


  //
  // Before we start, we have to set some global variables.
  //

  CreateUidTable();

  // Don't include per-node tracing.
  TraceLevel = tr_no_nodes;

  // We create our "program" node,
  // and then insert the graph (corresponding
  // to the top level of the program).

  ProgramRoot = new cProgram;

  // Remove the extension (starting from the period) from the filename.
  char * lastdot = strrchr (filename, '.');
  if (lastdot) {
    *lastdot = '\0';
  }
  ProgramRoot->Name = filename;

  ProgramRoot->UID = nextUID();

  ThisGraph = new cGraph;
  ThisGraph->Name = "main";
  ThisGraph->UID = ProgramRoot->UID;
  ProgramRoot->Graphs.Insert (ThisGraph);

  ThisUC = NULL;

  // Send compilation errors to stderr.
  ErrorLogFile = stderr;

  // Now parse the string.

  extern int yyMakeTree;

  yyset_input (buf, 0);
  yyMakeTree = 1;

  if (yyparse()) {
    // We had a parsing error.
    // Display the erroneous line.
    cerr << filename << ":" << yylinenum+1 << ": syntax error" << endl;
    // Find the line and display it.
    int lineno = 0;
    char * bufp = buf;
    while (*bufp && (lineno < yylinenum)) {
      if (*bufp == '\n') {
	lineno++;
      }
      bufp++;
    }
    // Print the line.
    while (*bufp) {
      printf ("%c", *bufp);
      if (*bufp == '\n') {
	break;
      }
      bufp++;
    }
    for (int i = 0; i < yycolumn; i++) {
      printf (" ");
    }
    printf ("^\n");
    exit (1);

  } else {

    int result = 1;

    // Parsing was successful.
    // Now we actually try to generate code
    // for the appropriate target.

    switch (TranslatorT) {
    case UNIFIED:
      result = Translate(DOUNIFIED);
      break;
    case SERIAL:
      result = Translate(DOSERIAL);
      break;
    case LIST:
      result = Translate(DOLISTING);
      break;
    case MPI:
      result = Translate(DOUNIFIED);
      break;
    case PVM:
      result = Translate(DOUNIFIED);
      break;
    default:
      assert (0);
      exit (1);
    }

    if (result == 1) {
      cout << filename << ": compilation successful." << endl;
    }

  }

  fclose (f);

  return 0;
}



